#include "llvm/ExecutionEngine/SectionMemoryManager.h"
#include "llvm/IR/LLVMContext.h"
#include "llvm/IR/Mangler.h"
#include "llvm/IRReader/IRReader.h"
#include "llvm/Object/ObjectFile.h"
#include "llvm/Support/CommandLine.h"
#include "llvm/Support/Debug.h"
#include "llvm/Support/DynamicLibrary.h"
#include "llvm/Support/Format.h"
#include "llvm/Support/SourceMgr.h"
#include "llvm/Support/TargetSelect.h"
#include "llvm/Target/TargetSubtargetInfo.h"

#include "llvm/ExecutionEngine/OrcJIT/CompileUtils.h"
#include "llvm/ExecutionEngine/OrcJIT/ObjectLinkingLayer.h"
#include "llvm/ExecutionEngine/OrcJIT/IRCompileLayer.h"
#include "llvm/ExecutionEngine/OrcJIT/LazyEmittingLayer.h"
#include "llvm/ExecutionEngine/OrcJIT/IndirectionUtils.h"
#include "llvm/ExecutionEngine/OrcJIT/CompileOnDemandLayer.h"

#include <memory>
#include <sstream>

using namespace llvm;

// Command line argument for IR file name.
cl::list<std::string> InputFiles(cl::desc("<input bitcodes>"),
                                cl::Positional, cl::ZeroOrMore);

typedef int (*MainFun)(int, const char*[]);

template <typename NativePtrT>
NativePtrT MakeNative(uint64_t P) {
  return reinterpret_cast<NativePtrT>(static_cast<uintptr_t>(P));
}

class MyJIT {
public:
  MyJIT()
    : TM(EngineBuilder().selectTarget()),
      Mang(TM->getSubtargetImpl()->getDataLayout()),
      CompileLayer(ObjectLayer, SimpleCompiler(*TM)),
      LazyEmitLayer(CompileLayer),
      ReentryLayer(LazyEmitLayer) { }

  void addModule(std::unique_ptr<Module> M) {
    std::vector<std::unique_ptr<Module>> S;
    S.push_back(std::move(M));
    ReentryLayer.addModuleSet(std::move(S),
                              llvm::make_unique<SectionMemoryManager>());
  }

  uint64_t getSymbolAddress(const std::string &Name) {
    std::string MangledName;
    {
      raw_string_ostream MangledNameStream(MangledName);
      Mang.getNameWithPrefix(MangledNameStream, Name);
    }
    return ReentryLayer.getSymbolAddress(MangledName, false);
  }

private:
  typedef ObjectLinkingLayer<> ObjLayerT;
  typedef IRCompileLayer<ObjLayerT> CompileLayerT;
  typedef LazyEmittingLayer<CompileLayerT> LazyEmitLayerT;
  typedef CompileOnDemandLayer<LazyEmitLayerT> ReentryLayerT;

  std::unique_ptr<TargetMachine> TM;
  Mangler Mang;

  ObjLayerT ObjectLayer;
  CompileLayerT CompileLayer;
  LazyEmitLayerT LazyEmitLayer;
  ReentryLayerT ReentryLayer;
};

// --- LAZY COMPILE TEST ---
int main(int argc, char* argv[]) {
  // Parse the command line options (this will initialize 'InputFiles').
  cl::ParseCommandLineOptions(argc, argv, "LLVM lazy JIT example.\n");

  InitializeNativeTarget();
  InitializeNativeTargetAsmPrinter();
  InitializeNativeTargetAsmParser();

#ifdef __APPLE__
  sys::DynamicLibrary::LoadLibraryPermanently(
      "/usr/lib/system/libsystem_c.dylib");
#endif

  LLVMContext &Context = getGlobalContext();
  SMDiagnostic Err;
  std::vector<std::unique_ptr<Module>> Modules;
  for (const auto& FileName : InputFiles)
    Modules.push_back(parseIRFile(FileName, Err, Context));

  MyJIT MJ;
  for (auto &M : Modules)
    MJ.addModule(std::move(M));

  MainFun Main = MakeNative<MainFun>(MJ.getSymbolAddress("main"));
  if (!Main) {
    llvm::errs() << "Bitcode does not contain a main function.\n";
    return 1;
  }

  int ArgC = 1;
  const char * ArgV[] = { InputFiles[0].c_str() };

  return Main(ArgC, ArgV);
}
